__author__ = """Nicolas Chiaruttini"""
__version__ = "0.1.0-SNAPSHOT"

# Inner classes, do not access and use them directly
# Used internally by the Abba package

# from abba_python.abba_private.AbbaAtlasNode import AbbaAtlasNode
