__author__ = """nicolas_chiaruttini"""
__version__ = "0.1.0-SNAPSHOT"

#from scijava_python_command.MagicJupyter import run_in_another_thread
#from scijava_python_command.ScijavaJupyterUI import enable_jupyter_ui
#from scijava_python_command.ScijavaCommand import ScijavaCommand, ScijavaInput, ScijavaOutput
